--------------------------------------------------------------------------
-- File vhdl/eu_inc/eu_inc.vhdl - INC Unit Execution
-- Copyright (C) 2002 Etienne Labarre (etienne.labarre@gadz.org)
--
-- Sun Jun 30 13:09:23 CEST 2002 : E. LABARRE
--  First version
--
-- Sat Jul 13 18:15:03 CEST 2002 : E. LABARRE
--  Unit written in sequential mode
--
-- Sun Jul 28 13:34:02 CEST 2002 : E. LABARRE
--  add operations : ABS, LSB0, LSB1
-- 
--------------------------BEGIN-VHDL-LICENCE-----------------------------
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
---------------------------END-VHDL-LICENCE------------------------------
--
-- SIMD MODES
-- see file find_lsb.vhdl
-- 
-- OPERATIONS MODE
-- eu_inc_mode operation test
-- 000      NEG       
-- 001      DEC       
-- 010      INC       
-- 011      ABS
-- 100      LSB0
-- 101      LSB1
--
--------------------------------------------------------------------------

-- include standard libraries
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.all;

-- include project libraries
LIBRARY work;
USE work.find_lsb;

-- Definition of EU_INC entity
entity eu_inc is
   generic ( eu_inc_width : natural := 6 );
   port (
      eu_inc_A    : in  std_ulogic_vector(2**eu_inc_width-1 downto 0);
      eu_inc_simd : in  std_ulogic_vector(eu_inc_width-4 downto 0);
      eu_inc_mode : in  std_ulogic_vector(2 downto 0);
      eu_inc_Y    : out std_ulogic_vector(2**eu_inc_width-1 downto 0)
      );
end eu_inc;

-- Description of EU_INC entity
architecture simple of eu_inc is
   type abs_array is array(eu_inc_width-3 downto 0) of std_ulogic_vector(2**eu_inc_width-1 downto 0);
   signal tree_in   : std_ulogic_vector(2**eu_inc_width-1 downto 0);
   signal tree_out  : std_ulogic_vector(2**eu_inc_width-1 downto 0);

   function abs_chunk( A, B : in std_ulogic_vector ) return std_ulogic_vector is
      variable L : natural := A'length;
      variable K : natural := A'high;
      variable sign : std_ulogic;

   begin
      sign := A(K);

      case sign is
         when '1' => return A;
	 when '0' => return B;
	 when others => return A;
      end case;
   end function;

begin
   -- Component find_lsb
   tree_find_lsb : entity find_lsb
      generic map( find_lsb_width => eu_inc_width )
      port map(
         find_lsb_input  => tree_in,
         find_lsb_output => tree_out,
         find_lsb_simd   => eu_inc_simd
         );

   eu_inc_in : process(eu_inc_A, eu_inc_simd, eu_inc_mode) is
   begin
     -- Input of unit
     case eu_inc_mode is 
        when "000"  => tree_in <= not eu_inc_A;
        when "001"  => tree_in <= not eu_inc_A;
        when "010"  => tree_in <= eu_inc_A;
        when "011"  => tree_in <= not eu_inc_A;
        when "100"  => tree_in <= eu_inc_A;
        when "101"  => tree_in <= not eu_inc_A;
        when others => tree_in <= eu_inc_A;
     end case;
   end process;

   eu_inc_out : process(tree_out, eu_inc_A, eu_inc_simd, eu_inc_mode) is
      variable L : natural := 2**eu_inc_width;
      variable N : natural;
      variable abs_simd : abs_array;
      variable tree_neg : std_ulogic_vector(2**eu_inc_width-1 downto 0);
      variable tree_inc : std_ulogic_vector(2**eu_inc_width-1 downto 0);
      variable tree_abs : std_ulogic_vector(2**eu_inc_width-1 downto 0);

   begin
      tree_neg := not tree_out xor eu_inc_A;
      tree_inc := tree_out xor eu_inc_A;

      -- perform abs tree
      for i in 0 to eu_inc_width-3 loop
         N := 2**i;
         for j in 0 to N-1 loop
	    abs_simd(i)((j+1)*L/N-1 downto j*L/N) := abs_chunk( 
	       eu_inc_A((j+1)*L/N-1 downto j*L/N), 
	       tree_neg((j+1)*L/N-1 downto j*L/N));
         end loop;
      end loop;

      -- search simd level and mux for abs
      tree_abs := abs_simd(0)(2**eu_inc_width-1 downto 0);
      for i in 0 to eu_inc_simd'high loop
         if eu_inc_simd(i) = '1' then
	    tree_abs := abs_simd(i+1)(2**eu_inc_width-1 downto 0);
	 end if;
      end loop;

      -- Output of unit
      case eu_inc_mode is
         when "000"  => eu_inc_Y <= tree_neg;
         when "001"  => eu_inc_Y <= tree_inc;
         when "010"  => eu_inc_Y <= tree_inc;
         when "011"  => eu_inc_Y <= tree_abs;
         when "100"  => eu_inc_Y <= tree_out;
         when "101"  => eu_inc_Y <= tree_out;
         when others => eu_inc_Y <= tree_inc;
      end case;
   end process;
end;
