--------------------------------------------------------------------------
-- f-cpu/vhdl/eu_inc/test_eu_cmp.vhdl - Testbench for the EU_CMP
-- Copyright (C) 2002 Etienne LABARRE (etienne.labarre@gadz.org)
-- 
-- Sun Jun 30 13:09:23 CEST 2002
--
--------------------------BEGIN-VHDL-LICENCE-----------------------------
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
---------------------------END-VHDL-LICENCE------------------------------
--
-- 
--------------------------------------------------------------------------

LIBRARY ieee;
    USE ieee.std_logic_1164.ALL;
    USE ieee.numeric_std.all;
-- text I/O
    use IEEE.std_logic_textio.all;
    use std.textio.all;
-- local functions
LIBRARY work;
    USE work.random.all;
    USE work.eu_cmp;

Entity test_eu_cmp64 is
  -- empty
end test_eu_cmp64;

Architecture testbench_eu_cmp of test_eu_cmp64 is

  constant n_test : natural := 10;

  constant WIDTH : natural := 6;
  constant UMX   : natural := 2**WIDTH;

  signal eu_cmp_in1_vector : std_ulogic_vector(UMX-1 downto 0);
  signal eu_cmp_in2_vector : std_ulogic_vector(UMX-1 downto 0);
  signal eu_cmp_out1_vector : std_ulogic_vector(UMX-1 downto 0);
  signal eu_cmp_out2_vector : std_ulogic_vector(UMX-1 downto 0);
  signal eu_cmp_ope        : std_ulogic_vector(2 downto 0);
  signal eu_cmp_simd_mode  : std_ulogic_vector(WIDTH-4 downto 0);

begin

   -- Component under test :
   eu_cmp_component : entity eu_cmp
      generic map (eu_cmp_width => WIDTH)
      port map(
	 eu_cmp_A    => eu_cmp_in1_vector,
	 eu_cmp_B    => eu_cmp_in2_vector,
	 eu_cmp_Y    => eu_cmp_out1_vector,
	 eu_cmp_Z    => eu_cmp_out2_vector,
	 eu_cmp_mode => eu_cmp_ope,
	 eu_cmp_simd => eu_cmp_simd_mode
	 );

   -- test bench :
   testbench : process is
      variable lout  : line;
      variable input1_vector    : std_ulogic_vector(UMX-1 downto 0);
      variable input2_vector    : std_ulogic_vector(UMX-1 downto 0);
      variable output1_vector   : std_ulogic_vector(UMX-1 downto 0);
      variable output2_vector   : std_ulogic_vector(UMX-1 downto 0);
      variable expected1_vector : std_ulogic_vector(UMX-1 downto 0);
      variable expected2_vector : std_ulogic_vector(UMX-1 downto 0);

      variable index      : integer;
      variable nb_chunk   : integer;
      variable nb_chunk1  : integer;
      variable base_chunk : integer;

      -- procedure for debug
      procedure debug_result1 is
         variable lout  : line;
      begin
         write(lout, string'("input 1  : "));
         write(lout, input1_vector);
	 writeline(output, lout);
         write(lout, string'("output 1 : "));
	 write(lout, output1_vector);
	 writeline(output, lout);
         write(lout, string'("expected 1 "));
	 write(lout, expected1_vector);
	 writeline(output, lout);
      end procedure;
      
      procedure debug_result2 is
         variable lout  : line;
      begin
         write(lout, string'("input 1  : "));
         write(lout, input1_vector);
	 writeline(output, lout);
         write(lout, string'("input 2  : "));
         write(lout, input2_vector);
	 writeline(output, lout);
         write(lout, string'("output 1 : "));
	 write(lout, output1_vector);
	 writeline(output, lout);
         write(lout, string'("output 2 : "));
	 write(lout, output2_vector);
	 writeline(output, lout);
         write(lout, string'("expected 1 "));
	 write(lout, expected1_vector);
	 writeline(output, lout);
         write(lout, string'("expected 2 "));
	 write(lout, expected2_vector);
	 writeline(output, lout);
      end procedure;

      -- function that return null vector
      function nul ( value : in std_ulogic_vector )
         return std_ulogic_vector is

         variable temp  : std_ulogic_vector(value'range);

      begin
         for i in value'high downto value'low loop
            temp(i) := '0';
         end loop;
         return temp;
      end function;

      -- function for cmp 2 values
      function cmp ( value1 : in std_ulogic_vector; value2 : in std_ulogic_vector )
         return std_ulogic_vector is

         variable temp  : std_ulogic_vector(value1'range);

      begin
         for i in value1'high downto value1'low loop
            temp(i) := value1(i) xor value2(i);
         end loop;
         return temp;
      end function;

      -- function for find MSB0
      function msb0 ( value : in std_ulogic_vector )
         return std_ulogic_vector is

         variable carry : std_ulogic;
         variable temp  : std_ulogic_vector(value'range);

      begin
         carry := '0';
         for i in value'high downto value'low loop
            temp(i) := carry;
            if carry = '0' then
               if value(i) = '0' then
                  carry := '1';
               end if;
            end if;
         end loop;
         return not temp;
      end function;

      -- function for find MSB1
      function msb1 ( value : in std_ulogic_vector )
         return std_ulogic_vector is

         variable carry : std_ulogic;
         variable temp  : std_ulogic_vector(value'range);

      begin
         carry := '0';
         for i in value'high downto value'low loop
            temp(i) := carry;
            if carry = '0' then
               if value(i) = '1' then
                  carry := '1';
               end if;
            end if;
         end loop;
         return not temp;
      end function;

      procedure test_CMP ( value : integer ) is

      begin
         for i in 0 to value loop
	    -- input = random number
            input1_vector := rand(input1_vector);
            input2_vector := rand(input2_vector);
	    -- apply to tested unit
	    eu_cmp_in1_vector <= input1_vector;
	    eu_cmp_in2_vector <= input2_vector;
	    -- search simd level
	    nb_chunk := 1;
	    for index in 0 to eu_cmp_simd_mode'length-1 loop
	       if eu_cmp_simd_mode(index) = '1' then
	          nb_chunk := nb_chunk*2;
	       end if;
	    end loop;
	    -- perform expected value
            for i in 1 to nb_chunk loop
	       expected1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk) := 
	          cmp(input1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk),
	              input2_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk));
            end loop;
	    -- wait for 1 cycle
	    wait for 1 ns;
	    -- Compare result
	    output1_vector := eu_cmp_out1_vector;
	    output2_vector := eu_cmp_out2_vector;
            debug_result2;
	 end loop;
      end procedure;
         
      procedure test_MSB0 ( value : integer ) is

      begin
         for i in 0 to value loop
	    -- input = random number
            input1_vector := rand(input1_vector);
	    -- apply to tested unit
	    eu_cmp_in1_vector <= input1_vector;
	    -- search simd level
	    nb_chunk := 1;
	    for index in 0 to eu_cmp_simd_mode'length-1 loop
	       if eu_cmp_simd_mode(index) = '1' then
	          nb_chunk := nb_chunk*2;
	       end if;
	    end loop;
	    -- perform expected value
            for i in 1 to nb_chunk loop
	       expected1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk) := 
	          msb0(input1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk));
            end loop;
	    -- wait for 1 cycle
	    wait for 1 ns;
	    -- Compare result
	    output1_vector := eu_cmp_out1_vector;
	    assert expected1_vector = output1_vector
	       report "bad msb0 perform"
	       severity warning;
            if output1_vector /= expected1_vector then
               debug_result1;
	    end if;
	 end loop;
      end procedure;
         
      procedure test_MSB1 ( value : integer ) is

      begin
         for i in 0 to value loop
	    -- input = random number
            input1_vector := rand(input1_vector);
	    -- apply to tested unit
	    eu_cmp_in1_vector <= input1_vector;
	    -- search simd level
	    nb_chunk := 1;
	    for index in 0 to eu_cmp_simd_mode'length-1 loop
	       if eu_cmp_simd_mode(index) = '1' then
	          nb_chunk := nb_chunk*2;
	       end if;
	    end loop;
	    -- perform expected value
            for i in 1 to nb_chunk loop
	       expected1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk) := 
	          msb1(input1_vector(i*UMX/nb_chunk-1 downto (i-1)*UMX/nb_chunk));
            end loop;
	    -- wait for 1 cycle
	    wait for 1 ns;
	    -- Compare result
	    output1_vector := eu_cmp_out1_vector;
	    assert expected1_vector = output1_vector
	       report "bad msb1 perform"
	       severity warning;
            if output1_vector /= expected1_vector then
               debug_result1;
	    end if;
	 end loop;
      end procedure;
         
  -- process test
  begin    
     
     write(lout, string'("******************* START OF TEST *******************"));
     writeline(output, lout);
    
     eu_cmp_simd_mode <= "000";
     wait for 1 ns;
     write(lout, string'("SIMD = "));
     write(lout, eu_cmp_simd_mode);
     writeline(output, lout);
     wait for 1 ns;

     write(lout, string'("Test of CMP"));
     writeline(output, lout);
     eu_cmp_ope <= "000";
     test_CMP(n_test);

     write(lout, string'("Test of MSB0"));
     writeline(output, lout);
     eu_cmp_ope <= "100";
     test_MSB0(n_test);

     write(lout, string'("Test of MSB1"));
     writeline(output, lout);
     eu_cmp_ope <= "101";
     test_MSB1(n_test);

     eu_cmp_simd_mode <= "001";
     wait for 1 ns;
     write(lout, string'("SIMD = "));
     write(lout, eu_cmp_simd_mode);
     writeline(output, lout);
     wait for 1 ns;

     write(lout, string'("Test of CMP"));
     writeline(output, lout);
     eu_cmp_ope <= "000";
     test_CMP(n_test);

     write(lout, string'("Test of MSB0"));
     writeline(output, lout);
     eu_cmp_ope <= "100";
     test_MSB0(n_test);

     write(lout, string'("Test of MSB1"));
     writeline(output, lout);
     eu_cmp_ope <= "101";
     test_MSB1(n_test);

     eu_cmp_simd_mode <= "011";
     wait for 1 ns;
     write(lout, string'("SIMD = "));
     write(lout, eu_cmp_simd_mode);
     writeline(output, lout);
     wait for 1 ns;

     write(lout, string'("Test of CMP"));
     writeline(output, lout);
     eu_cmp_ope <= "000";
     test_CMP(n_test);

     write(lout, string'("Test of MSB0"));
     writeline(output, lout);
     eu_cmp_ope <= "100";
     test_MSB0(n_test);

     write(lout, string'("Test of MSB1"));
     writeline(output, lout);
     eu_cmp_ope <= "101";
     test_MSB1(n_test);

     eu_cmp_simd_mode <= "111";
     wait for 1 ns;
     write(lout, string'("SIMD = "));
     write(lout, eu_cmp_simd_mode);
     writeline(output, lout);
     wait for 1 ns;

     write(lout, string'("Test of CMP"));
     writeline(output, lout);
     eu_cmp_ope <= "000";
     test_CMP(n_test);

     write(lout, string'("Test of MSB0"));
     writeline(output, lout);
     eu_cmp_ope <= "100";
     test_MSB0(n_test);

     write(lout, string'("Test of MSB1"));
     writeline(output, lout);
     eu_cmp_ope <= "101";
     test_MSB1(n_test);

     write(lout, string'("*******************  END OF TEST ********************"));
     writeline(output, lout);
     wait;
  end process;

end; 
